Imports System.Math

Public Class Form1

    ' The last point clicked.
    Private PointClicked As PointF = New PointF(-1, -1)

    ' Redraw on resize.
    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Me.ResizeRedraw = True
    End Sub

    ' Save the clicked point.
    Private Sub Form1_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        ' Save the point.
        PointClicked = New PointF(e.X, e.Y)

        ' Redraw.
        Me.Invalidate()
    End Sub

    ' Draw the circle, point, and tangent lines.
    Private Sub Form1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        ' Draw the circle.
        Dim radius As Single = Math.Min( _
            Me.ClientSize.Width / 6.0F, _
            Me.ClientSize.Height / 6.0F)
        Dim cx As Single = Me.ClientSize.Width - radius - 20
        Dim cy As Single = Me.ClientSize.Height - radius - 20
        e.Graphics.FillEllipse(Brushes.LightBlue, _
            cx - radius, cy - radius, _
            2 * radius, 2 * radius)
        e.Graphics.DrawEllipse(Pens.Blue, _
            cx - radius, cy - radius, _
            2 * radius, 2 * radius)

        If (PointClicked.X < 0) Then Return

        ' Find the tangents.
        Dim pt1, pt2 As PointF
        If FindTangents(New PointF(cx, cy), radius, PointClicked, pt1, pt2) _
        Then
            e.Graphics.DrawLine(Pens.Green, PointClicked, pt1)
            e.Graphics.DrawLine(Pens.Green, PointClicked, pt2)
        End If

        ' Draw the point.
        e.Graphics.FillEllipse(Brushes.Pink, _
            PointClicked.X - 3, PointClicked.Y - 3, _
            7, 7)
        e.Graphics.DrawEllipse(Pens.Red, _
            PointClicked.X - 3, PointClicked.Y - 3, _
            7, 7)
    End Sub

    ' Find the tangent points for this circle and external point.
    ' Return true if we find the tangents, false if the point is
    ' inside the circle.
    Private Function FindTangents(ByVal center As PointF, ByVal radius As Single, _
        ByVal external_point As PointF, ByRef pt1 As PointF, ByRef pt2 As PointF) As Boolean

        ' Find the distance squared from the
        ' external point to the circle's center.
        Dim dx As Double = center.X - external_point.X
        Dim dy As Double = center.Y - external_point.Y
        Dim D_squared As Double = dx * dx + dy * dy
        If (D_squared < radius * radius) Then
            pt1 = New PointF(-1, -1)
            pt2 = New PointF(-1, -1)
            Return False
        End If

        ' Find the distance from the external point
        ' to the tangent points.
        Dim L As Double = Sqrt(D_squared - radius * radius)

        ' Find the points of intersection between
        ' the original circle and the circle with
        ' center external_point and radius dist.
        FindCircleCircleIntersections( _
            center.X, center.Y, radius, _
            external_point.X, external_point.Y, CSng(L), _
            pt1, pt2)

        Return True
    End Function

    ' Find the points where the two circles intersect.
    Private Function FindCircleCircleIntersections( _
     ByVal cx0 As Single, ByVal cy0 As Single, ByVal radius0 As _
         Single, _
     ByVal cx1 As Single, ByVal cy1 As Single, ByVal radius1 As _
         Single, _
     ByRef intersection1 As PointF, ByRef intersection2 As  _
         PointF) As Integer
        ' Find the distance between the centers.
        Dim dx As Single = cx0 - cx1
        Dim dy As Single = cy0 - cy1
        Dim dist As Double = Math.Sqrt(dx * dx + dy * dy)

        ' See how many solutions there are.
        If (dist > radius0 + radius1) Then
            ' No solutions, the circles are too far apart.
            intersection1 = New PointF(Single.NaN, Single.NaN)
            intersection2 = New PointF(Single.NaN, Single.NaN)
            Return 0
        ElseIf (dist < Math.Abs(radius0 - radius1)) Then
            ' No solutions, one circle contains the other.
            intersection1 = New PointF(Single.NaN, Single.NaN)
            intersection2 = New PointF(Single.NaN, Single.NaN)
            Return 0
        ElseIf ((dist = 0) AndAlso (radius0 = radius1)) Then
            ' No solutions, the circles coincide.
            intersection1 = New PointF(Single.NaN, Single.NaN)
            intersection2 = New PointF(Single.NaN, Single.NaN)
            Return 0
        Else
            ' Find a and h.
            Dim a As Double = (radius0 * radius0 - _
                    radius1 * radius1 + dist * dist) / (2 * _
                        dist)
            Dim h As Double = Math.Sqrt(radius0 * radius0 - a * _
                a)

            ' Find P2.
            Dim cx2 As Double = cx0 + a * (cx1 - cx0) / dist
            Dim cy2 As Double = cy0 + a * (cy1 - cy0) / dist

            ' Get the points P3.
            intersection1 = New PointF( _
                CSng(cx2 + h * (cy1 - cy0) / dist), _
                CSng(cy2 - h * (cx1 - cx0) / dist))
            intersection2 = New PointF( _
                CSng(cx2 - h * (cy1 - cy0) / dist), _
                CSng(cy2 + h * (cx1 - cx0) / dist))

            ' See if we have 1 or 2 solutions.
            If (dist = radius0 + radius1) Then Return 1
            Return 2
        End If
    End Function
End Class
